/* SPDX-License-Identifier: GPL-2.0 WITH Linux-syscall-note */
#ifndef __ASM_MIPS_BITSPERLONG_H
#define __ASM_MIPS_BITSPERLONG_H

#define __BITS_PER_LONG _MIPS_SZLONG

#include <asm-generic/bitsperlong.h>

#endif /* __ASM_MIPS_BITSPERLONG_H */
