/* SPDX-License-Identifier: GPL-2.0 WITH Linux-syscall-note */
#ifndef _ASM_LOONGARCH_PERF_REGS_H
#define _ASM_LOONGARCH_PERF_REGS_H

enum perf_event_loongarch_regs {
	PERF_REG_LOONGARCH_PC,
	PERF_REG_LOONGARCH_R1,
	PERF_REG_LOONGARCH_R2,
	PERF_REG_LOONGARCH_R3,
	PERF_REG_LOONGARCH_R4,
	PERF_REG_LOONGARCH_R5,
	PERF_REG_LOONGARCH_R6,
	PERF_REG_LOONGARCH_R7,
	PERF_REG_LOONGARCH_R8,
	PERF_REG_LOONGARCH_R9,
	PERF_REG_LOONGARCH_R10,
	PERF_REG_LOONGARCH_R11,
	PERF_REG_LOONGARCH_R12,
	PERF_REG_LOONGARCH_R13,
	PERF_REG_LOONGARCH_R14,
	PERF_REG_LOONGARCH_R15,
	PERF_REG_LOONGARCH_R16,
	PERF_REG_LOONGARCH_R17,
	PERF_REG_LOONGARCH_R18,
	PERF_REG_LOONGARCH_R19,
	PERF_REG_LOONGARCH_R20,
	PERF_REG_LOONGARCH_R21,
	PERF_REG_LOONGARCH_R22,
	PERF_REG_LOONGARCH_R23,
	PERF_REG_LOONGARCH_R24,
	PERF_REG_LOONGARCH_R25,
	PERF_REG_LOONGARCH_R26,
	PERF_REG_LOONGARCH_R27,
	PERF_REG_LOONGARCH_R28,
	PERF_REG_LOONGARCH_R29,
	PERF_REG_LOONGARCH_R30,
	PERF_REG_LOONGARCH_R31,
	PERF_REG_LOONGARCH_MAX,
};
#endif /* _ASM_LOONGARCH_PERF_REGS_H */
