// SPDX-License-Identifier: LGPL-2.1
// Copyright (C) 2018, 2019 Red Hat Inc, Arnaldo Carvalho de Melo <acme@redhat.com>
//
#ifndef __PERF_GET_CURRENT_DIR_NAME_H
#ifndef HAVE_GET_CURRENT_DIR_NAME
char *get_current_dir_name(void);
#endif // HAVE_GET_CURRENT_DIR_NAME
#endif // __PERF_GET_CURRENT_DIR_NAME_H
