/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1995, 1999, 2001, 2002 by Ralf Baechle
 */
#ifndef _ASM_ERRNO_H
#define _ASM_ERRNO_H

#include <uapi/asm/errno.h>


/* The biggest error number defined here or in <linux/errno.h>. */
#define EMAXERRNO	1133

#endif /* _ASM_ERRNO_H */
